# Copyright (c) 2013, 2014 Erik Hetzner
# 
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from decimal import Decimal
from ofxparse.ofxparse import Transaction, InvestmentTransaction
from ledgerautosync import EmptyInstitutionException

AUTOSYNC_INITIAL = "autosync_initial"
ALL_AUTOSYNC_INITIAL = "all.%s"%(AUTOSYNC_INITIAL)

def clean_ofx_id(ofxid):
    ofxid = ofxid.replace('/', '_')
    ofxid = ofxid.replace('$', '_')
    ofxid = ofxid.replace(' ', '_')
    return ofxid

class Formatter(object):
    def __init__(self, account, name, indent=4, ledger=None, fid=None):
        self.acctid=account.account_id
        if fid is not None:
            self.fid = fid
        else:
            if account.institution is None:
                raise EmptyInstitutionException("Institution provided by OFX is empty and no fid supplied!")
            else:
                self.fid=account.institution.fid
        self.name = name
        self.lgr = ledger
        self.indent = indent
        self.currency = account.statement.currency
        self.currency = self.currency.upper()
        if self.currency == "USD": self.currency = "$"

    def mk_ofxid(self, txnid):
        return clean_ofx_id("%s.%s.%s"%(self.fid, self.acctid, txnid))

    def mk_dynamic_account(self, txn, exclude):
        if self.lgr is None:
            return "Expenses:Misc"
        else:
            payee = self.format_payee(txn)
            account = self.lgr.get_account_by_payee(payee, exclude)
            if account is None:
                return "Expenses:Misc"
            else:
                return account

    def format_amount(self, amount, reverse=False, unlimited=False):
        if unlimited:
            amt = str(abs(amount))
        else:
            amt = "%0.2f"%(abs(amount))
        if amount.is_signed() != reverse:
            return "-%s%s"%(self.currency, amt)
        else:
            return "%s%s"%(self.currency, amt)

    def format_payee(self, txn):
        payee = None
        memo = None
        if (hasattr(txn, 'payee')): payee = txn.payee
        if (hasattr(txn, 'memo')): memo = txn.memo
        
        if (payee is None) and (memo is None):
            return "UNKNOWN"
        if (payee is None) or txn.memo.startswith(payee):
            return memo
        elif (memo is None) or payee.startswith(memo):
            return payee
        else:
            return "%s %s"%(payee, memo)

    def format_date(self, date):
        return date.strftime("%Y/%m/%d")

    def format_balance(self, statement):
        retval = ""
        if (hasattr(statement, 'balance_date')):
            date = statement.balance_date
        else:
            date = statement.end_date
        if (hasattr(statement, 'balance')):
            retval += "%s * --Autosync Balance Assertion\n"%(self.format_date(date))
            retval += self.format_txn_line(self.name, self.format_amount(Decimal("0")), " = %s"%(self.format_amount(statement.balance)))
        return retval

    def format_initial_balance(self, statement):
        retval = ""
        if (hasattr(statement, 'balance')):
            initbal = statement.balance
            for txn in statement.transactions:
                initbal -= txn.amount
            retval += "%s * --Autosync Initial Balance\n"%(self.format_date(statement.start_date))
            retval += "%s; ofxid: %s\n"%(" "*self.indent, self.mk_ofxid(AUTOSYNC_INITIAL))
            retval += self.format_txn_line(self.name, self.format_amount(initbal))
            retval += self.format_txn_line("Assets:Equity", self.format_amount(initbal, reverse=True))
        return retval

    def format_txn_line(self, acct, amt, suffix=""):
        space_count = 52 - self.indent - len(acct) - len(amt)
        if space_count < 2:
            space_count = 2
        return "%s%s%s%s%s\n"%(" "*self.indent, acct, " "*space_count, amt, suffix)

    def format_txn(self, txn):
        retval = ""
        ofxid = self.mk_ofxid(txn.id)
        if isinstance(txn, Transaction):
            retval += "%s %s\n"%(self.format_date(txn.date), self.format_payee(txn))
            retval += "%s; ofxid: %s\n"%(" "*self.indent, ofxid)
            retval += self.format_txn_line(self.name, self.format_amount(txn.amount))
            retval += self.format_txn_line(self.mk_dynamic_account(txn, exclude=self.name), self.format_amount(txn.amount, reverse=True))
        elif isinstance(txn, InvestmentTransaction):
            if txn.settleDate is not None:
                retval = "%s=%s %s\n"%(txn.tradeDate.strftime("%Y/%m/%d"), txn.settleDate.strftime("%Y/%m/%d"), txn.memo)
            else:
                retval = "%s %s\n"%(txn.tradeDate.strftime("%Y/%m/%d"), self.format_payee(txn))
            retval += "%s; ofxid: %s\n"%(" "*self.indent, ofxid)
            retval += self.format_txn_line(acct=self.name, amt=str(txn.units), 
                                           suffix=" %s @ %s"%(txn.security, self.format_amount(txn.unit_price, unlimited=True)))
            retval += self.format_txn_line(self.name, self.format_amount(txn.units * txn.unit_price, reverse=True))
        return retval



# Copyright (c) 2013, 2014 Erik Hetzner
# 
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from ledgerautosync.cli import import_ofx
from ledgerautosync.ledgerwrap import Ledger, HLedger, LedgerPython
from ledgerautosync import EmptyInstitutionException
import os.path

from unittest import TestCase
from nose.tools import raises

class TestCli(TestCase):
    def setUp(self):
        self.ledgers = [LedgerPython(os.path.join('fixtures', 'empty.lgr')),
                        HLedger(os.path.join('fixtures', 'empty.lgr')),
                        Ledger(os.path.join('fixtures', 'empty.lgr'), no_pipe=True)]

    @raises(EmptyInstitutionException)
    def test_no_institution_no_fid(self):
        for lgr in self.ledgers:
            ofxpath = os.path.join('fixtures', 'no-institution.ofx')
            import_ofx(lgr, ofxpath, accountname="Assets:Foo")

    def test_no_institution(self):
        for lgr in self.ledgers:
            ofxpath = os.path.join('fixtures', 'no-institution.ofx')
            import_ofx(lgr, ofxpath, accountname="Assets:Foo", fid=1234567890)

    @raises(EmptyInstitutionException)
    def test_no_institution_no_accountname(self):
        for lgr in self.ledgers:
            ofxpath = os.path.join('fixtures', 'no-institution.ofx')
            import_ofx(lgr, ofxpath, fid=1234567890)

    def test_apostrophe(self):
        for lgr in self.ledgers:
            ofxpath = os.path.join('fixtures', 'apostrophe.ofx')
            import_ofx(lgr, ofxpath, fid=1234567890)



# Copyright (c) 2013, 2014 Erik Hetzner
# 
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from ledgerautosync.formatter import Formatter
from ledgerautosync.ledgerwrap import Ledger
import os.path

from ofxparse import OfxParser

from unittest import TestCase

class TestFormatter(TestCase):
    def test_checking(self):
        ofx = OfxParser.parse(file(os.path.join('fixtures', 'checking.ofx')))
        formatter = Formatter(account=ofx.account, name="Foo", indent=2)
        self.assertEqual(formatter.format_txn(ofx.account.statement.transactions[0]),
"""2011/03/31 DIVIDEND EARNED FOR PERIOD OF 03/01/2011 THROUGH 03/31/2011 ANNUAL PERCENTAGE YIELD EARNED IS 0.05%
  ; ofxid: 1101.1452687~7.0000486
  Foo                                          $0.01
  Expenses:Misc                               -$0.01
""")
        self.assertEqual(formatter.format_txn(ofx.account.statement.transactions[1]),
"""2011/04/05 AUTOMATIC WITHDRAWAL, ELECTRIC BILL WEB(S )
  ; ofxid: 1101.1452687~7.0000487
  Foo                                        -$34.51
  Expenses:Misc                               $34.51
""")
        self.assertEqual(formatter.format_txn(ofx.account.statement.transactions[2]),
"""2011/04/07 RETURNED CHECK FEE, CHECK # 319 FOR $45.33 ON 04/07/11
  ; ofxid: 1101.1452687~7.0000488
  Foo                                        -$25.00
  Expenses:Misc                               $25.00
""")

    def test_indent(self):
        ofx = OfxParser.parse(file(os.path.join('fixtures', 'checking.ofx')))
        formatter = Formatter(account=ofx.account, name="Foo", indent=4)
        self.assertEqual(formatter.format_txn(ofx.account.statement.transactions[0]),
"""2011/03/31 DIVIDEND EARNED FOR PERIOD OF 03/01/2011 THROUGH 03/31/2011 ANNUAL PERCENTAGE YIELD EARNED IS 0.05%
    ; ofxid: 1101.1452687~7.0000486
    Foo                                        $0.01
    Expenses:Misc                             -$0.01
""")

    def test_investments(self):
        ofx = OfxParser.parse(file(os.path.join('fixtures', 'fidelity.ofx')))
        formatter = Formatter(account=ofx.account, name="Foo", indent=2)
        self.assertEqual(formatter.format_txn(ofx.account.statement.transactions[0]),
"""2012/07/20 YOU BOUGHT
  ; ofxid: 7776.01234567890.0123456789020201120120720
  Foo                                      100.00000 458140100 @ $25.635000000
  Foo                                      -$2563.50
""")
        # test no payee/memo
        self.assertEqual(formatter.format_txn(ofx.account.statement.transactions[1]),
"""2012/07/27 UNKNOWN
  ; ofxid: 7776.01234567890.0123456789020901120120727
  Foo                                      128.00000 G7945E105 @ $39.390900000
  Foo                                      -$5042.04
""")

    def test_dynamic_account(self):
        ofx = OfxParser.parse(file(os.path.join('fixtures', 'checking.ofx')))
        ledger = Ledger(os.path.join('fixtures', 'checking-dynamic-account.lgr'))
        formatter = Formatter(account=ofx.account, name="Assets:Foo", ledger=ledger, indent=2)
        self.assertEqual(formatter.format_txn(ofx.account.statement.transactions[1]),
"""2011/04/05 AUTOMATIC WITHDRAWAL, ELECTRIC BILL WEB(S )
  ; ofxid: 1101.1452687~7.0000487
  Assets:Foo                                 -$34.51
  Expenses:Bar                                $34.51
""")

    def test_balance_assertion(self):
        ofx = OfxParser.parse(file(os.path.join('fixtures', 'checking.ofx')))
        ledger = Ledger(os.path.join('fixtures', 'checking.lgr'))
        formatter = Formatter(account=ofx.account, name="Assets:Foo", ledger=ledger)
        self.assertEqual(formatter.format_balance(ofx.account.statement),
"""2013/05/25 * --Autosync Balance Assertion
    Assets:Foo                                 $0.00 = $100.99
""")

    def test_initial_balance(self):
        ofx = OfxParser.parse(file(os.path.join('fixtures', 'checking.ofx')))
        ledger = Ledger(os.path.join('fixtures', 'checking.lgr'))
        formatter = Formatter(account=ofx.account, name="Assets:Foo", ledger=ledger)
        self.assertEqual(formatter.format_initial_balance(ofx.account.statement),
"""2000/01/01 * --Autosync Initial Balance
    ; ofxid: 1101.1452687~7.autosync_initial
    Assets:Foo                               $160.49
    Assets:Equity                           -$160.49
""")



# Copyright (c) 2013, 2014 Erik Hetzner
# 
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
import xml.etree.ElementTree as ET
import os
import re
import subprocess
from subprocess import Popen, PIPE
from threading import Thread
from Queue import Queue, Empty
from ledgerautosync.formatter import clean_ofx_id
import logging

def hledger_clean(a):
    def clean_str(s):
        s = s.replace('(', '\(')
        s = s.replace(')', '\)')
        return s
    return [ clean_str(s) for s in a ]

def pipe_clean(a):
    def clean_str(s):
        s = s.replace('/', '\\\\/')
        s = s.replace('%', '')
        if not(re.match(r"^\w+$", s)):
            s = "\"%s\""%(s)
        return s
    return [ clean_str(s) for s in a ]

def windows_clean(a):
    def clean_str(s):
        s = s.replace('%', '')
        s = s.replace(' ', '\ ')
        s = s.replace('/', '\/')
        return s
    return [ clean_str(s) for s in a ]

def clean_payee(s):
    s = s.replace('%', '')
    s = s.replace('/', '\/')
    s = s.replace("'", "")
    return s

def all_or_none(seq):
    """Returns the first value of seq if all values of seq are equal, or returns None."""
    if len(seq) == 0: 
        return None
    def f(x,y):
        if (x == y): return x
        else: return None
    return reduce(f, seq, seq[0])

def mk_ledger(ledger_file):
    try:
        import ledger
        return LedgerPython(ledger_file, string_read=False)
    except ImportError:
        if os.name == 'posix':
            if ((subprocess.call("which ledger > /dev/null", shell=True) == 0) and
                (Popen(["ledger", "--version"], stdout=PIPE).communicate()[0]).startswith("Ledger 3")):
                return Ledger(ledger_file)
            elif subprocess.call("which hledger > /dev/null", shell=True) == 0:
                return HLedger(ledger_file)
            else:
                raise Exception("Neither ledger 3 nor hledger found!")
        else:
            # windows, I guess ... just assume ledger
            return Ledger(ledger_file)

class Ledger(object):
    def __init__(self, ledger_file=None, no_pipe=True):
        self._item = ""
        def enqueue_output(out, queue):
            buff = ""
            while (buff != None):
                buff = out.read(1)
                if (buff != None):
                    self._item += buff
                if self._item.endswith("] "): # prompt
                    queue.put(self._item[0:-2])
                    self._item = ""
            out.close()
        self.use_pipe = (os.name == 'posix') and not(no_pipe)
        self.args = ["ledger"]
        if ledger_file is not None:
            self.args += ["-f", ledger_file]
        if self.use_pipe:
            self.p = Popen(self.args, bufsize=1, stdin=PIPE, stdout=PIPE,
                           close_fds=True)
            self.q = Queue()
            self.t = Thread(target=enqueue_output, args=(self.p.stdout, self.q))
            self.t.daemon = True # thread dies with the program
            self.t.start()
            # read output until prompt
            try:
                self.q.get(True, 5)
            except Empty:
                logging.error("Could not get prompt (]) from ledger!")
                logging.error("Received: %s"%(self._item))
                exit(1)

    def run(self, cmd):
        if self.use_pipe:
            self.p.stdin.write("xml ")
            self.p.stdin.write(" ".join(pipe_clean(cmd)))
            self.p.stdin.write("\n")
            logging.debug(" ".join(pipe_clean(cmd)))
            try:
                return ET.fromstring(self.q.get(True, 5))
            except Empty:
                logging.error("Could not get prompt from ledger!")
                exit(1)
        else:
            cmd = self.args + ["xml"] + cmd
            if os.name == 'nt':
                cmd = windows_clean(cmd)
            return ET.fromstring(subprocess.check_output(cmd))
            
    def get_transaction(self, q):
        d = self.run(q).findall('.//transactions/transaction')
        if len(d) == 0:
            return None
        else:
            return d[0]

    def check_transaction_by_ofxid(self, ofxid):
        return (self.get_transaction(["-E", "meta", "ofxid=%s"%(clean_ofx_id(ofxid))]) != None)
        
    def get_account_by_payee(self, payee, exclude):
        try:
            txn = self.run(["--real", "payee", clean_payee(payee)])
            if txn is None: return None
            else: 
                accts = [ node.text for node in txn.findall('.//transactions/transaction/postings/posting/account/name') ]
                accts_filtered = [ a for a in accts if a != exclude ]
                if accts_filtered: return accts_filtered[-1]
                else: return None
        except:
            logging.error("Error checking --real payee for %s"%(clean_payee(payee)))

class LedgerPython(object):
    def __init__(self, ledger_file=None, string_read=True):
        # sanity check for ledger python interface
        try:
            import ledger
        except ImportError:
            raise Exception("Ledger python interface not found!")
        if ledger_file is None:
            # TODO - better loading
            raise Exception
        else:
            if string_read:
                self.session = ledger.Session()
                self.journal = self.session.read_journal_from_string(open(ledger_file).read())
            else:
                self.journal = ledger.read_journal(ledger_file)

    def check_transaction_by_ofxid(self, ofxid):
        q = self.journal.query("-E meta ofxid=\"%s\""%(clean_ofx_id(ofxid)))
        return len(q) > 0

    def get_account_by_payee(self, payee, exclude):
        q  = self.journal.query("--real payee '%s'"%(clean_payee(payee)))
        accts = [ p.account for p in q ]
        accts_filtered = [ a for a in accts if a != exclude ]
        if accts_filtered: return str(accts_filtered[-1])
        else: return None

class HLedger(object):
    def __init__(self, ledger_file=None):
        self.args = ["hledger"]
        if ledger_file is not None:
            self.args += ["-f", ledger_file]

    def run(self, cmd):
        cmd = hledger_clean(self.args + cmd)
        if os.name == 'nt':
            cmd = windows_clean(cmd)
        logging.debug(" ".join(cmd))
        return subprocess.check_output(cmd)

    def check_transaction_by_ofxid(self, ofxid):
        cmd = ["reg", "tag:ofxid=%s"%(clean_ofx_id(ofxid))]
        return self.run(cmd) != ''

    def get_account_by_payee(self, payee, exclude):
        cmd = ["reg", "-w200", "desc:%s"%(payee)]
        lines = self.run(cmd).splitlines()
        accts = [ l[92:172].strip() for l in lines ]
        accts_filtered = [ a for a in accts if a != exclude ]
        if accts_filtered: return accts_filtered[-1]
        else: return None



# Copyright (c) 2013, 2014 Erik Hetzner
# 
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from ofxparse import OfxParser
import logging

class Synchronizer(object):
    def __init__(self, lgr):
        self.lgr = lgr

    def parse_file(self, path, accountname=None):
        ofx = OfxParser.parse(file(path))
        return (ofx, self.filter(ofx))

    def is_txn_synced(self, acctid, txn):
        ofxid = "%s.%s"%(acctid, txn.id)
        return self.lgr.check_transaction_by_ofxid(ofxid)
    
    def filter(self, ofx):
        txns = ofx.account.statement.transactions
        if len(txns) == 0:
            sorted_txns = txns
        elif hasattr(txns[0], 'settleDate'):
            sorted_txns = sorted(txns, key=lambda t: t.settleDate)
        else:
            sorted_txns = sorted(txns, key=lambda t: t.date)
        acctid = ofx.account.account_id
        return [ txn for txn in sorted_txns if not(self.is_txn_synced(acctid, txn)) ]

    def get_new_txns(self, acct, max_days=999999, resync=False):
        if resync or (max_days < 7):
            days = max_days
        else:
            days = 7
        last_txns_len = 0
        while (True):
            logging.debug("Downloading %d days of transactions for %s (max_days=%d)."%(days, acct.description, max_days))
            raw = acct.download(days)
            ofx = OfxParser.parse(raw)
            if not(hasattr(ofx, 'account')):
                # some banks return this for no txns
                if (days >= max_days):
                    logging.debug("Hit max days.")
                    # return None to let the caller know that we don't
                    # even have account info
                    return (None, None)
                else:
                    days = days * 2
                    if (days > max_days): days = max_days
                    logging.debug("empty account: increasing days ago to %d."%(days))
                    last_txns_len = 0
            else:
                txns = ofx.account.statement.transactions
                new_txns = self.filter(ofx)
                logging.debug("txns: %d"%(len(txns)))
                logging.debug("new txns: %d"%(len(new_txns)))
                if ((len(txns) > 0) and (last_txns_len == len(txns))):
                    # not getting more txns than last time; we have
                    # reached the beginning
                    logging.debug("Not getting more txns than last time, done.")
                    return (ofx, new_txns)
                elif (len(txns) > len(new_txns)) or (days >= max_days):
                    # got more txns than were new or hit max_days, we've
                    # reached a stopping point
                    if (days >= max_days):
                        logging.debug("Hit max days.")
                    else:
                        logging.debug("Got some stale txns.")
                    return (ofx, new_txns)
                else:
                    # all txns were new, increase how far back we go
                    days = days * 2
                    if (days > max_days): days = max_days
                    logging.debug("Increasing days ago to %d."%(days))
                    last_txns_len = len(txns)



# Copyright (c) 2013, 2014 Erik Hetzner
# 
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
import os
import os.path
from ledgerautosync.ledgerwrap import Ledger, HLedger, LedgerPython

from unittest import TestCase

class TestLedger(TestCase):
    def setUp(self):
        self.ledgers = [LedgerPython(os.path.join('fixtures', 'checking.lgr')),
                        HLedger(os.path.join('fixtures', 'checking.lgr')),
                        Ledger(os.path.join('fixtures', 'checking.lgr'), no_pipe=True)]

    def test_transaction(self):
        for lgr in self.ledgers:
            self.assertTrue(lgr.check_transaction_by_ofxid("1101.1452687~7.0000486"))
        
#    def test_transaction_substring(self):
#        self.assertIsNone(lgr.get_transaction_by_ofxid("1452687~7"))

    def test_nonexistent_transaction(self):
        for lgr in self.ledgers:
            self.assertFalse(lgr.check_transaction_by_ofxid("FOO"))

    def test_empty_transaction(self):
        for lgr in self.ledgers:
            self.assertTrue(lgr.check_transaction_by_ofxid("empty"))
    
    def test_get_account_by_payee(self):
        for lgr in self.ledgers:
            account = lgr.get_account_by_payee("AUTOMATIC WITHDRAWAL, ELECTRIC BILL WEB(S )", exclude="Assets:Foo")
            self.assertEqual(account, "Expenses:Bar", msg="%s != Expenses:Bar with %s"%(account, lgr))

    def test_get_ambiguous_account_by_payee(self):
        ledgers = [LedgerPython(os.path.join('fixtures', 'checking-dynamic-account.lgr')),
                   Ledger(os.path.join('fixtures', 'checking-dynamic-account.lgr'), no_pipe=True),
                   HLedger(os.path.join('fixtures', 'checking-dynamic-account.lgr'))]
        
        for lgr in ledgers:
            account = lgr.get_account_by_payee("Generic", exclude="Assets:Foo")
            # shoud use the latest
            self.assertEqual(account, "Expenses:Bar")
                                                                                                             
    def test_ofx_quoting(self):
        for lgr in self.ledgers:
            self.assertEqual(lgr.check_transaction_by_ofxid("1/2"), True, msg="Did not find 1/2 in %s"%(lgr))
        
    # broken in ledger
#    def test_multiple_transaction(self):
#        lgr = Ledger("fixtures/multiple.lgr")
#        txn = lgr.get_transaction_by_payee("Baz")
#        self.assertEqual(len(txn) > 1)



# Copyright (c) 2013, 2014 Erik Hetzner
# 
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from ledgerautosync.cli import sync
from ledgerautosync.ledgerwrap import Ledger, LedgerPython, HLedger
from ofxclient.config import OfxConfig
import os.path

from unittest import TestCase
from mock import Mock

class TestCli(TestCase):
    def test_run(self):
        for lgr in [LedgerPython(os.path.join('fixtures', 'empty.lgr')),
                    HLedger(os.path.join('fixtures', 'empty.lgr')),
                    Ledger(os.path.join('fixtures', 'empty.lgr'), no_pipe=True)]:
            config = OfxConfig(os.path.join('fixtures', 'ofxclient.ini'))
            acct = config.accounts()[0]
            acct.download = Mock(return_value=file(os.path.join('fixtures', 'checking.ofx')))
            config.accounts = Mock(return_value=[acct])
            sync(lgr, config, 7)

    def test_empty_run(self):
        for lgr in [LedgerPython(os.path.join('fixtures', 'checking.lgr')),
                    HLedger(os.path.join('fixtures', 'checking.lgr')),
                    Ledger(os.path.join('fixtures', 'checking.lgr'), no_pipe=True)]:
            config = OfxConfig(os.path.join('fixtures', 'ofxclient.ini'))
            acct = config.accounts()[0]
            acct.download = Mock(return_value=file(os.path.join('fixtures', 'checking.ofx')))
            config.accounts = Mock(return_value=[acct])
            sync(lgr, config, 7)




#!/usr/bin/env python

# Copyright (c) 2013, 2014 Erik Hetzner
# 
# This file is part of ledger-autosync
#
# ledger-autosync is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# ledger-autosync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ledger-autosync. If not, see
# <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from ofxclient.config import OfxConfig
import argparse
from ledgerautosync import EmptyInstitutionException
from ledgerautosync.formatter import Formatter, AUTOSYNC_INITIAL, ALL_AUTOSYNC_INITIAL
from ledgerautosync.sync import Synchronizer
from ledgerautosync.ledgerwrap import mk_ledger, Ledger, HLedger
import logging
import re
import sys
import traceback
import os, os.path

def find_ledger_file():
    """Returns main ledger file path or raise exceptio if it cannot be found."""
    ledgerrcpath = os.path.abspath(os.path.expanduser("~/.ledgerrc"))
    if os.environ.has_key("LEDGER_FILE"):
        return os.path.abspath(os.path.expanduser(os.environ["LEDGER_FILE"]))
    elif os.path.exists(ledgerrcpath):
        # hacky
        ledgerrc = open(ledgerrcpath).read()
        return os.path.abspath(os.path.expanduser(re.match(r".*--file\s+([^\s]+).*", ledgerrc).group(1)))
    else:
        raise Exception("LEDGER_FILE environment variable not set, and no .ledgerrc file found, and -l argument no supplied.")

def maybe_print_initial(ofx, ledger, formatter):
    if (not(ledger.check_transaction_by_ofxid(formatter.mk_ofxid(AUTOSYNC_INITIAL))) and
        not(ledger.check_transaction_by_ofxid(ALL_AUTOSYNC_INITIAL))):
        print formatter.format_initial_balance(ofx.account.statement)

def sync(ledger, config, max_days=90, resync=False, indent=4, initial=False, assertions=False):
    sync = Synchronizer(ledger)
    for acct in config.accounts():
        try:
            (ofx, txns) = sync.get_new_txns(acct, resync=resync, max_days=max_days)
            if ofx is not None:
                formatter = Formatter(account=ofx.account, name=acct.description, ledger=ledger, indent=indent)
                if initial:
                    maybe_print_initial(ofx, ledger, formatter)
                for txn in txns:
                    print formatter.format_txn(txn)
                if assertions:
                    print formatter.format_balance(ofx.account.statement)
        except KeyboardInterrupt:
            raise
        except:
            sys.stderr.write("Caught exception processing %s"%(acct.description))
            traceback.print_exc(file=sys.stderr)
                             
def import_ofx(ledger, path, accountname=None, indent=4, initial=False, assertions=False, fid=None):
    sync = Synchronizer(ledger)
    (ofx, txns) = sync.parse_file(path)
    if accountname is None:
        if ofx.account.institution is not None:
            accountname = "%s:%s"%(ofx.account.institution.organization, ofx.account.account_id)
        else:
            raise EmptyInstitutionException("Institution provided by OFX is empty and no accountname supplied!")
    formatter = Formatter(account=ofx.account, name=accountname, ledger=ledger, indent=indent, fid=fid)
    if initial:
        maybe_print_initial(ofx, ledger, formatter)
    for txn in txns:
        print formatter.format_txn(txn)
    if assertions:
        print formatter.format_balance(ofx.account.statement)
        
def run(args=None):
    if args is None: args = sys.argv[1:]

    parser = argparse.ArgumentParser(description='Synchronize ledger.')
    parser.add_argument('-m', '--max', type=int, default=90,
                        help='maximum number of days to process')
    parser.add_argument('-r', '--resync', action='store_true', default=False,
                        help='do not stop until max days reached')
    parser.add_argument('PATH', nargs='?', help='do not sync; import from OFX file')
    parser.add_argument('-a', '--account', type=str, default=None,
                        help='set account name for import')
    parser.add_argument('-l', '--ledger', type=str, default=None,
                        help='specify ledger file to READ for syncing')
    parser.add_argument('-i', '--indent', type=int, default=4,
                        help='number of spaces to use for indentation')
    parser.add_argument('--initial', action='store_true', default=False,
                        help='create initial balance entries')
    parser.add_argument('--fid', type=int, default=None,
                        help='pass in fid value for OFX files that do not supply it')
    parser.add_argument('--assertions', action='store_true', default=False,
                        help='create balance assertion entries')
    parser.add_argument('-d', '--debug', action='store_true', default=False,
                        help='enable debug logging')
    parser.add_argument('--hledger', action='store_true', default=False,
                        help='force use of hledger (on by default if invoked as hledger-autosync)')
    parser.add_argument('--slow', action='store_true', default=False,
                        help='use slow, but possibly more robust, method of calling ledger (no subprocess)')
    args = parser.parse_args(args)
    if sys.argv[0][-16:] == "hledger-autosync":
      args.hledger = True

    ledger_file = None
    if args.ledger:
        ledger_file = args.ledger
    else:
        ledger_file = find_ledger_file()
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    if args.hledger:
        ledger = HLedger(ledger_file)
    elif args.slow:
        ledger = Ledger(ledger_file=ledger_file, no_pipe=True)
    else:
        ledger = mk_ledger(ledger_file)
    if args.PATH is None:
        config_dir = os.environ.get('XDG_CONFIG_HOME', os.path.join(os.path.expanduser("~"), '.config'))
        config_file = os.path.join(config_dir, 'ofxclient.ini')
        if (os.path.exists(config_file)):
            config = OfxConfig(file_name=config_file)
        else:
            config = OfxConfig()
        sync(ledger, config, max_days=args.max, resync=args.resync, indent=args.indent, initial=args.initial, assertions=args.assertions)
    else:
        import_ofx(ledger, args.PATH, args.account, indent=args.indent, initial=args.initial, assertions=args.assertions, fid=args.fid)

if __name__ == '__main__':
    run()
